@@ -11,6 +11,7 @@ from lxml import etree
 import base64
 import unicodedata
 import zlib
+import codecs
 from io import BytesIO
 from collections import OrderedDict
 
@@ -25,20 +26,40 @@ class PayloadChecksumError(Exception):
 class DynamicDict(Adapter):
     """ListContainer <---> Container
     Convenience mapping so we dont have to iterate ListContainer to find
-    the right item"""
+    the right item
 
-    def __init__(self, key, subcon):
+    FIXME: lump kwarg was added to get around the fact that InnerHeader is
+    not truly a dict.  We lump all 'binary' InnerHeaderItems into a single list
+    """
+
+    def __init__(self, key, subcon, lump=[]):
         super(DynamicDict, self).__init__(subcon)
         self.key = key
+        self.lump = lump
 
     # map ListContainer to Container
     def _decode(self, obj, context, path):
-        d = OrderedDict((item[self.key], item) for item in obj)
+        d = OrderedDict()
+        for l in self.lump:
+            d[l] = ListContainer([])
+        for item in obj:
+            if item[self.key] in self.lump:
+                d[item[self.key]].append(item)
+            else:
+                d[item[self.key]] = item
+
         return Container(d)
 
     # map Container to ListContainer
     def _encode(self, obj, context, path):
-        return ListContainer(obj.values())
+        l = []
+        for key in obj:
+            if key in self.lump:
+                l += obj[key]
+            else:
+                l.append(obj[key])
+
+        return ListContainer(l)
 
 def Reparsed(subcon_out):
     class Reparsed(Adapter):
@@ -98,14 +119,21 @@ def compute_key_composite(password=None, keyfile=None):
             try:
                 with open(keyfile, 'rb') as f:
                     key = f.read()
+
+                    try:
+                        int(key, 16)
+                        is_hex = True
+                    except ValueError:
+                        is_hex = False
                     # if the length is 32 bytes we assume it is the key
                     if len(key) == 32:
                         keyfile_composite = key
                     # if the length is 64 bytes we assume the key is hex encoded
-                    if len(key) == 64:
-                        keyfile_composite =  key.decode('hex')
+                    elif len(key) == 64 and is_hex:
+                        keyfile_composite =  codecs.decode(key, 'hex')
                     # anything else may be a file to hash for the key
-                    keyfile_composite = hashlib.sha256(key).digest()
+                    else:
+                        keyfile_composite = hashlib.sha256(key).digest()
             except:
                 raise IOError('Could not read keyfile')
 
